{***************************************************}
{*  EXEID v0.01                                   *}
{* Sample program for EXE identification           *}
{* Copyright (C) 1997 by Brczi, Gbor             *}
{* E-mail: sting@crnl-gimn.crnl.hu                 *}
{* Donated to the public domain.                   *}
{***************************************************}
program EXEID;

uses Dos,Objects;

type
     { Standard DOS EXE header }
     TEXEHeader = record
       EXESign              : word;    { 'MZ' }
       LastPageUsage        : word;    { number of bytes used on last 512 byte page               }
       TotalPages           : word;    { total num of 512-byte pages in executable                }
       RelocEntries         : word;    { number of relocation entries                             }
       HeaderSize           : word;    { header size in paragraph                                 }
       MinAddMemSize        : word;    { min paragraphs of memory needed in addition to exe size  }
       MaxAddMemSize        : word;    { max paragraphs to allocate in addition to exe size       }
       InitSS               : word;    { initial SS relative to start of executable               }
       InitSP               : word;    { initial SP                                               }
       CheckSum             : word;    { one's complement of sum of all words in executable       }
       EntryPoint           : longint; { inital CS:IP relative to start of executable             }
       RelocTableOfs        : word;    { offset within header of relocation table                 }
       OverlayNo            : word;    { 0000h = main program                                     }
       case integer of
       0 : { New Executables }
          ( Unknown          : longint;
            Behavior         : word;
            AddBehavior      : array[1..26] of byte; { reserved for additional behavior info                            }
            NEHeaderOfs      : longint;              { offset of NE header within disk file ; 00000000h if plain MZ exe }
          );
     end;

     { Windows & OS/2 New Executable Header }
     TNEHeader = record
       NESignature               : word;
       LinkerVersion             : word;
       EntryTableOfs             : word;
       EntryTableSize            : word;
       CRC                       : longint;
       ProgramFlags              : byte;
       ApplicationFlags          : byte;
       AutoDSIndex               : word;
       InitLocalHeapSize         : word;
       InitStackSize             : word;
       EntryPoint                : pointer;
       InitSS_SP                 : pointer;
       SegmentCount              : word;
       ModuleRefCount            : word;
       NonResidentNameTabLen     : word;
       SegmentTableOfs           : word;
       ResourceTableOfs          : word;
       ResidentNameTableOfs      : word;
       ModuleRefernceTableOfs    : word;
       ImportedNamesTabOfs       : word;
       NonResidentNameTableOfs   : word;
       MoveableEntryPointCount   : word;
       FileAlignmentShiftCount   : word;
       ResourceTableEntriesCount : word;
       TargetOS                  : byte;
       EXEFlags2                 : byte;
       ThunkRetOfs               : word;
       SegrefThunksOfs           : word;
       MinCodeSwapSize           : word;
       MinWindowsVersion         : word;
     end;

     { Windows & OS/2 Linear Executable Header }
     TLEHeader = record
       LESignature               : word;
       ByteOrder                 : byte;
       WordOrder                 : byte;
       FileFormatLevel           : longint;
       CPUType                   : word;
       TargetOS                  : word;
       ModuleVersion             : word;
       ModuleType                : word;
       { .... }
     end;

const
      { Standard DOS EXE header size }
      EXEHeaderSize = 28;

      { New Executable application flag bitfield masks & values }
      neafApplicationTypeMask    = 1+2+4;
      neafDLLorDriver            = 128;

      neafFullScreen             =  1;
      neafWindowsAPICompatible   =  2;
      neafWindowsAPIRequired     =  3;

      { Target OS value constants }
      osUnknown                  =  0;
      osOS2                      =  1;
      osWindows                  =  2;
      osEDOS4                    =  3;
      osWin386                   =  4;
      osBOSS                     =  5;

procedure Help;
begin
  writeln('Usage: EXEID <filespec1> [<filespec2> ..]');
  writeln('Displays information about executable files.');
  Halt(1);
end;

const
    ftBinary            = 0;
    ftEXE               = 1;
    ftEXE_NE            = 2;
    ftEXE_PE            = 3;
    ftEXE_LE            = 4;
    ftEXE_W3            = 5;
    ftEXE_LX            = 6;

type
    TFileInformation = record
       Typ        : byte;
       Size       : longint;
       case byte of
         ftBinary    : ();
         ftEXE       : (HeaderSize : longint;
                        ImageSize  : longint;
                        OvrSize    : longint;
                        Overlay    : boolean;
                        case integer of
                          ftEXE    : (
                                     MinMem   : longint;
                                     );
                          ftEXE_NE : (
                                     OS             : byte;
                                     MinWinVersion  : word;
                                     IsDLL          : boolean;
                                     );
                          ftEXE_PE,
                          ftEXE_LE,
                          ftEXE_LX   : ();
                          ftEXE_W3   : ();
                       );
     end;

function IntToStr(L: longint): string;
var S: string;
begin
  Str(L,S);
  IntToStr:=S;
end;

function IntToStrZ(L: longint; MinLen: byte): string;
var S: string;
begin
  Str(L,S);
  while length(S)<MinLen do S:='0'+S;
  IntToStrZ:=S;
end;

function UpcaseStr(S: string): string;
var I: integer;
begin
  for I:=1 to length(S) do S[I]:=Upcase(S[I]);
  UpcaseStr:=S;
end;

function MemToStr(var Buf; Len: byte): string;
var S: string;
begin
  S[0]:=chr(Len); if Len>0 then Move(Buf,S[1],Len);
  MemToStr:=S;
end;

function VersionToStr(V: word): string;
var S: string[10];
begin
  S:=IntToStr(V div 256)+'.'+IntToStrZ(V mod 256,2);
  VersionToStr:=S;
end;

procedure WriteFileInfo(FI: TFileInformation);
var S: string;
begin
  writeln('File size       : ',FI.Size);
  case FI.Typ of
       ftBinary : S:='binary (COM)';
       ftEXE    : S:='executable (EXE)';
       ftEXE_NE : S:='Windows New Executable (NE)';
       ftEXE_PE : S:='Windows Portable Executable (PE)';
       ftEXE_LE : S:='Windows Linear Executable (LE)';
       ftEXE_LX : S:='OS/2 Linear Executable (LX)';
       ftEXE_W3 : S:='Windows WIN386.EXE (W3)';
  end;
  writeln('File type       : ',S);
  if FI.Typ in[ftEXE..ftEXE_W3] then
     begin
  writeln('EXE information : ','header ',FI.HeaderSize,' bytes, image ',FI.ImageSize,
          ' bytes, overlay ',FI.OvrSize,' bytes');
     if FI.Typ=ftEXE then
     begin
  writeln('Memory required : approx. ',FI.MinMem div 1024,' KB');
     end else
     if FI.Typ in[ftEXE_NE,ftEXE_LE] then
     begin
      case FI.OS of
        osOS2         : S:='OS/2';
        osWindows     : begin
                          S:='Windows ';
                          if FI.Typ=ftEXE_NE then S:=S+' '+VersionToStr(FI.MinWinVersion);
                        end;
        osEDOS4       : S:='European MS-DOS';
        osWindows     : begin
                          S:='Windows 386';
                          if FI.Typ=ftEXE_NE then S:=S+' '+VersionToStr(FI.MinWinVersion);
                        end;
        osBOSS        : S:='BOSS (Borland Operating System Services)';
      else S:='???';
      end;
  writeln('Target OS       : ',S);
     end;
     end;
end;

procedure ProcessFile(Path: PathStr);
var S         : TBufStream;
    Buf       : array[0..4095] of byte;
    EH        : TEXEHeader absolute Buf;
    NEH       : TNEHeader absolute Buf;
    LEH       : TLEHeader absolute Buf;
    FI        : TFileInformation;
    TotalSize : longint;
    S2        : string;
begin
  writeln('Processing file : ',Path);
  S.Init(Path,stOpenRead,4096);
  if S.Status<>stOk then writeln('Error opening file ',Path) else
  begin
    FillChar(FI,SizeOf(FI),0);
    FI.Size:=S.GetSize;
    if FI.Size<EXEHeaderSize then FI.Typ:=ftBinary else
    begin
      S.Read(EH,SizeOf(EH));
      if S.Status<>stOk then writeln('Error reading file ',Path) else
      begin
        if MemToStr(EH.EXESign,2)<>'MZ' then FI.Typ:=ftBinary else
        begin
          FI.Typ:=ftEXE;
          TotalSize:=longint(EH.TotalPages-1)*512+EH.LastPageUsage;
          FI.HeaderSize:=longint(EH.HeaderSize)*16;
          FI.ImageSize:=TotalSize-FI.HeaderSize;
          FI.OvrSize:=FI.Size-TotalSize;
          FI.Overlay:=EH.OverlayNo<>0;
          FI.MinMem:=FI.ImageSize+longint(EH.MinAddMemSize)*16;
          if FI.Size>SizeOf(EH) then
          begin
            S.Seek(0);
            S.Read(EH,SizeOf(EH));
            if EH.NEHeaderOfs<>0 then
            begin
            S.Seek(EH.NEHeaderOfs);
            S.Read(NEH,SizeOf(NEH));
            S2:=MemToStr(NEH.NESignature,2);
            if S2='NE' then FI.Typ:=ftEXE_NE else
            if S2='PE' then FI.Typ:=ftEXE_PE else
            if S2='LE' then FI.Typ:=ftEXE_LE else
            if S2='LX' then FI.Typ:=ftEXE_LX else
            if S2='W3' then FI.Typ:=ftEXE_W3;
            if FI.Typ=ftEXE_NE then
            begin
              FI.OS:=NEH.TargetOS;
              FI.MinWinVersion:=NEH.MinWindowsVersion;
            end;
            if FI.Typ=ftEXE_LE then
            begin
              FI.OS:=LEH.TargetOS;
            end;
            end;
          end;
          WriteFileInfo(FI);
        end;
      end;
    end;
  end;
  S.Done;
end;

function GetDirOf(P: PathStr): string;
var D: DirStr;
    E: ExtStr;
    N: NameStr;
begin
  FSplit(P,D,N,E);
  if copy(D,length(D),1)<>'\' then D:=D+'\';
  GetDirOf:=D;
end;

procedure ProcessFileSpec(S: PathStr);
var SR: SearchRec;
    I : integer;
begin
  if (copy(S,length(S),1)='\') or (copy(S,length(S),1)=':') then S:=S+'*.*';
  if copy(S,1,2)<>'\\' then S:=FExpand(S);
  I:=0;
  FindFirst(S,AnyFile and not Directory,SR);
  while DosError=0 do
        begin
          Inc(I);
          ProcessFile(GetDirOf(S)+SR.Name);
          FindNext(SR);
        end;
  if I=0 then writeln('Could not find any file matching ',S);
end;

procedure ProcessFileSpecs;
var I: integer;
begin
  I:=1;
  repeat
    ProcessFileSpec(UpcaseStr(ParamStr(I)));
    Inc(I);
  until I>ParamCount;
end;

BEGIN
  writeln(' EXEID v.09  Copyright (C) 1997 by Brczi Gbor');
  if (ParamCount=0) or (ParamStr(1)='?') or (ParamStr(1)='/?') or (ParamStr(1)='/HELP')
     then Help;
  ProcessFileSpecs;
END.