{***************************************************}
{*  MEMCB v0.9                                    *}
{* DOS Memory Block listing program in BP 7.0      *}
{* Copyright (C) 1997 by Brczi, Gbor             *}
{* E-mail: sting@crnl-gimn.crnl.hu                 *}
{* Donated to the public domain.                   *}
{***************************************************}
{$M 8192,0,0}
uses Dos,Crt;

type
      { DOS MCB memory managment record structure (undocumented) }
      PMCB = ^TMCB;
      TMCB = record
        Typ       : char;
        OwnerPSP  : word;
        Size      : word;
        Reserved  : array[1..3] of byte;
        Name      : array[1..8] of char;
        Data      : record end;
      end;

      { DOS internal buffer (DOS data segment) managment record structure (undocumented) }
      PDataSegmentMCB = ^TDataSegmentMCB;
      TDataSegmentMCB = record
        ID        : char;
        Next      : word;
        Size      : word;
        Reserved  : array[1..3] of byte;
        Name      : array[1..8] of char;
      end;

      { DOS Program Segment Prefix }
      PPSP = ^TPSP;
      TPSP = record
        INT20h         : word;
        MemTop         : word;
        Reserved       : word;
        Avail          : word;
        Reserved2      : word;
        Terminate      : pointer;
        CtrlBreak      : pointer;
        CriticalError  : pointer;
        Reserved3      : array[1..22] of byte;
        EnvSeg         : word;
        Reserved4      : array[1..209] of byte;
      end;

const LineCount: integer = 0;

procedure writeln(S: string);
{ Procedure with simple screen-page-managment }
begin
  system.writeln(S);
  Inc(LineCount);
  if LineCount=24 then
     begin
       system.write('Press any key to continue...');
       readkey;
       write(#13); ClrEol;
       LineCount:=0;
     end;
end;

function IntToStrL(L: longint; MinLen: byte): string;
{ Converts longint number to string }
var S: string;
begin
  Str(L:MinLen,S);
  IntToStrL:=S;
end;

function RealToStrL(R: real; MinLen,Decimals: byte): string;
{ Converts real number to string }
var S: string;
begin
  Str(R:0:Decimals,S);
  while length(S)<MinLen do S:=' '+S;
  RealToStrL:=S;
end;

function WordToHex(W: word): string;
{ Converts word value to hex }
var S: string;
    I: integer;
const Digits : array[0..15] of char = '0123456789ABCDEF';
begin
  S:='';
  for I:=3 downto 0 do
      S:=S+Digits[(W div trunc(exp(I*ln(16)))) mod 16];
  WordToHex:=S;
end;

function ASCIIZToStr(var A; MaxLen: byte): string;
{ Convert null-terminated (ASCIIZ) string to Pascal string }
var S: string;
begin
  Move(A,S[1],MaxLen);
  S[0]:=#0;
  while (length(S)<MaxLen) and (S[length(S)+1]>=#32) do
        S:=S+S[length(S)+1];
  ASCIIZToStr:=S;
end;

function RExpand(S: string; MinLen: byte): string;
{ Expands string to a minimum length with padding spaces at right }
begin
  while length(S)<MinLen do S:=S+' ';
  RExpand:=S;
end;

function GetRootMCBAddr: PMCB; assembler;
{ Get address of root (first) MCB (undocumented) }
asm
  mov   ah, 52h
  int   21h
  mov   dx, word ptr es:[bx-2]
  xor   ax, ax
end;

const
      { Special owner-PSP values used by DOS and 386MAX }
      psp_FREE         = $0000;
      psp_XMSUMB       = $0006;
      psp_XMSHole      = $0007;
      psp_DOS          = $0008;
      psp_386MAX       = $FFF7;
      psp_UMBCB        = $FFFA;
      psp_Excluded     = $FFFD;
      psp_UMB          = $FFFE;
      psp_386MAXDevice = $FFFF;

      { Memory block type constants }
      mbt_Free               =   0;
      mbt_Buffers            =   1;
      mbt_BuffersEMS         =   2;
      mbt_Device             =   3;
      mbt_DeviceAppendage    =   4;
      mbt_Files              =   5;
      mbt_IFS                =   6;
      mbt_LastDrive          =   7;
      mbt_Stacks             =   8;
      mbt_FCBS               =   9;
      mbt_Program            =  10;
      mbt_Unknown            =  11;
      mbt_DOS                =  12;
      mbt_Environment        =  13;
      mbt_Data               =  14;
      mbt_UMB                =  15;
      mbt_Excluded           =  16;
      mbt_386MAX             =  17;
      mbt_XMSHole            =  18;

      { Display strings for memory block type constants }
      Typs : array[0..18] of string[20] =
        ( 'Free space', 'Buffers', 'Buffers (EMS)', 'Device', 'Device appendage',
          'Files', 'IFS', 'LastDrive', 'Stacks', 'FCBS', 'Program', 'Unknown',
          'DOS','Environment','Data','UMB','Excluded','386MAX','XMS hole');

      { Maximum number of memory blocks displayed (can be increased up to 2800) }
      MaxBlockCount = 200;

type
     { Internal structure to hold information on one memory block }
     TMemoryBlock = record
       MCBAddr  : PMCB;       { address of original MCB                 }
       PSP      : PPSP;       { pointer to PSP of MCB's owner           }
       Typ      : byte;       { type of memory block (see mbt_XXXX)     }
       Owner    : string[8];  { owner of memory block                   }
       Size     : longint;    { size of memory block in bytes           }
       HighNext : boolean;    { is next block located in high memory    }
     end;

var BlockCount : integer;
    Blocks     : array[1..MaxBlockCount] of TMemoryBlock;

procedure CollectBlocks;
{ Collects and completes information about memory blocks }
var MCB    : PMCB;            { current MCB accessed                    }
    NextMCB: PMCB;            { next MCB                                }
    DSMCB  : PDataSegmentMCB; { current DOS data-segment block accessed }
    I,J    : integer;         { general 'for' variables                 }
    HighMem: boolean;         { currently scanning high memory blocks   }
function FindFirstHighMCB: PMCB;
{ Locates first high memory block with brute force scanning }
var MCB: PMCB;
    OK : boolean;
    Next: PMCB;
begin
  MCB:=ptr($9FFF,0); OK:=false;
  repeat
    MCB:=ptr(seg(MCB^)+1,0);
    if (MCB^.Typ in['M','Z']) and (MCB^.OwnerPSP<>0) then
       begin
         Next:=ptr(seg(MCB^)+MCB^.Size,0);
         OK:=Next^.Typ in['M','Z'];
       end;
  until (OK=true) or (seg(MCB^)<$9FFF);
  if OK=false then MCB:=nil;
  FindFirstHighMCB:=MCB;
end;
begin
  FillChar(Blocks,SizeOf(Blocks),0);
  BlockCount:=0; HighMem:=false;
  MCB:=GetRootMCBAddr;
  while (MCB<>nil) and (BlockCount<MaxBlockCount) do
  with Blocks[BlockCount+1] do
        begin
          Inc(BlockCount);
          MCBAddr:=MCB;
          Size:=longint(MCB^.Size)*16;
          if MCB^.Typ='M'
             then NextMCB:=Ptr(seg(MCB^)+1+(Size div 16),0)
             else NextMCB:=nil;
          if (NextMCB=nil) and (HighMem=false) then
             begin
               HighMem:=true;
               NextMCB:=FindFirstHighMCB;
             end;
          Owner:=''; Typ:=mbt_Unknown;
          PSP:=ptr(MCB^.OwnerPSP,0);
          case MCB^.OwnerPSP of
               psp_Free         : begin Typ:=mbt_Free; Owner:=''; end;
               psp_XMSUMB       : Typ:=mbt_UMB;
               psp_XMSHole      : Typ:=mbt_XMSHole;
               psp_386MAX       : Typ:=mbt_386MAX;
               psp_UMBCB        : Typ:=mbt_UMB;
               psp_Excluded     : Typ:=mbt_Excluded;
               psp_UMB          : Typ:=mbt_UMB;
               psp_386MAXDevice : begin
                                    Typ:=mbt_Device;
                                    Owner:=ASCIIZToStr(MCB^.Name,8);
                                  end;
               psp_DOS  : begin
                            DSMCB:=@MCB^.Data;
                            if BlockCount=1 then
                            repeat
                              with Blocks[BlockCount] do
                              begin
                              Typ:=mbt_DOS;
                              Owner:=ASCIIZToStr(DSMCB^.Name,8);
                              case DSMCB^.ID of
                                   'B' : begin
                                           Typ:=mbt_Buffers;
                                           Owner:='System';
                                         end;
                                   'C' : begin
                                           Typ:=mbt_BuffersEMS;
                                           Owner:='System';
                                         end;
                                   'D' : Typ:=mbt_Device;
                                   'E' : begin
                                           Typ:=mbt_DeviceAppendage;
                                           Owner:='System';
                                         end;
                                   'F' : begin
                                           Typ:=mbt_Files;
                                           Owner:='System';
                                         end;
                                   'I' : Typ:=mbt_IFS;
                                   'L' : begin
                                           Typ:=mbt_LastDrive;
                                           Owner:='System';
                                         end;
                                   'S' : begin
                                           Typ:=mbt_Stacks;
                                           Owner:='System';
                                         end;
                                   'X' : begin
                                           Typ:=mbt_FCBS;
                                           Owner:='System';
                                         end;
                              end;
                              if Owner='' then Owner:='DOS';
                              Owner:=RExpand(Owner,8);
                              end;
                              DSMCB:=ptr(DSMCB^.Next+DSMCB^.Size,0);
                              if DSMCB<>pointer(NextMCB) then Inc(BlockCount);
                            until DSMCB=pointer(NextMCB);
                          end;
          else begin
                 Owner:=ASCIIZToStr(MCB^.Name,8);
                 if Size=0
                    then begin
                           Typ:=mbt_Free;
                           Owner:='';
                         end
                    else begin
                           if (PSP^.INT20h=$20CD) or (PSP^.INT20h=$27CD)
                              then begin
                                     if MCB^.OwnerPSP=seg(MCB^)+1
                                        then Typ:=mbt_Program
                                        else if PSP^.EnvSeg=seg(MCBAddr^)+1
                                                then Typ:=mbt_Environment
                                                else Typ:=mbt_Data;
                                   end
                              else begin
                                    if MCB^.OwnerPSP=seg(MCB^)+1
                                       then Typ:=mbt_Device
                                       else if PSP^.EnvSeg=seg(MCB^)+1
                                               then Typ:=mbt_Environment
                                               else Typ:=mbt_Data;

                                   end;
                         end;
               end; { case .. else .. }
          end; { case MCB^.OwnerPSP of .. }
          Owner:=RExpand(Owner,8);
          MCB:=NextMCB;
          HighNext:=HighMem;
          if (MCB<>nil) and (BlockCount=MaxBlockCount) then
             writeln('!!! To many blocks (>200) to display. List is incomplete. !!!');
        end;
  for I:=1 to BlockCount do
  for J:=1 to BlockCount do
      begin
        if Blocks[I].Typ in[mbt_Program,mbt_Device] then
           if (Blocks[I].PSP=Blocks[J].PSP) then
           begin
           if
           ((Blocks[J].Typ in[mbt_Program,mbt_Device,mbt_DOS,mbt_IFS,mbt_DeviceAppendage,
                          mbt_Files,mbt_FCBS,mbt_Buffers,mbt_BuffersEMS,mbt_LastDrive,
                          mbt_Stacks,mbt_Free])=false) then
            begin
              Blocks[J].Owner:=Blocks[I].Owner;
            end; end else
            begin
              if ((Blocks[I].Typ in [mbt_DOS,mbt_IFS,mbt_DeviceAppendage,mbt_Files,
                                    mbt_FCBS,mbt_BuffersEMS,mbt_Buffers,
                                    mbt_LastDrive,mbt_Stacks,mbt_Free])=false)
                  then
                  begin
                     if BlockS[J].Typ in [mbt_Program, mbt_Device] then
                        begin
                          if (Blocks[J].PSP = Blocks[I].PSP) then
                             begin
                               Blocks[I].Owner := Blocks[J].Owner;
                               if (Blocks[J].PSP^.EnvSeg <> 0) and
                                  (seg(Blocks[I].MCBAddr^)+1 = Blocks[J].PSP^.EnvSeg)
                                  then begin
                                         Blocks[I].Typ := mbt_Environment;
                                       end
                                  else begin
                                         Blocks[I].Typ := mbt_Data;
                                       end;
                             end;
                        end;
                  end;
            end;
      end;
end;

var I: integer;
    H: boolean;

BEGIN
  writeln(' MEMCB  Copyright (C) 1997 by Brczi Gbor');
  writeln('Lists DOS memory control blocks');
  I:=DOSVersion;
  if (I<$0400) and (I>=$0100) then begin writeln('Requires DOS 4.0 or later.'); Halt(1); end;
  writeln('');
  writeln(' Address           Size          Owner     Type ');
  writeln(' -------   -------------------  --------  --------------');
  CollectBlocks; H:=false;
  for I:=1 to BlockCount do with Blocks[I] do
      begin
          writeln('   '+WordToHex(seg(MCBAddr^))+'    '+IntToStrL(Size,8)+' ('+RealToStrL(Size/1024,6,1)+'K)  '+
                  Owner+'   '+Typs[Typ]);
        if I<>BlockCount then
        if H<>HighNext then writeln('--------------------- High memory ---------------------');
        H:=HighNext;
      end;
END.