(*****************************************************************************
*                                                                            *
*                                 D  S  P                                    *
*                                                                            *
*****************************************************************************)

unit dsp;

interface

 type
      { Egy 64K szles bjttmb tpus definilsa        }
      memory = array[0..65534] of byte;
      { Mutat definilsa a 64K szles bjttmb szmra }
      memptr = ^memory;

 const
      { A SoundBlaster krtya rendszerinformcii }
      BaseAddr   : word = $220; { SB I/O bziscm       }
      IRQnumber  : byte =  $05; { SB megszaktsi vonal }
      DMAchannel : byte =  $01; { SB DMA csatornaszm   }

      { DSP parancsok }

      SetSampleRate = $40; { A mintavtelezsi id belltsa (idkonstans) }

      { A DSP kimenetvel kapcsolatos parancsok }
      SpeakerOn     = $d1; { A kimenet bekapcsolsa            }
      SpeakerOff    = $d3; { A kimenet kikapcsolsa            }
      SpeakerStatus = $d8; { A kimenet llapotnak lekrdezse }

      { Direkt md lejtszs s felvtel }
      DirectDAC     = $10; { Direkt lejtszs (D/A konverzi)  }
      DirectADC     = $20; { Direkt felvtel  (A/D konverzi)  }

      { A DMA tvitellel kapcsolatos parancsok }
      DSP_DMA_stop  = $d0; { A DMA tvitel felfggesztse      }
      DSP_DMA_cont  = $d4; { A DMA tvitel folytatsa          }

      NormalDMA_DAC = $14; { Norml 8 bites lejtszs (D/A konverzi) }
      NormalDMA_ADC = $24; { Norml 8 bites felvtel  (A/D konverzi) }

      { Impulzuskd-modullt DMA tviteli parancsok }
      ADPCM4_DMA    = $74; { 4 bites ADPCM                     }
      ADPCM4r_DMA   = $75; { 4 bites ADPCM referencival       }
      ADPCM26_DMA   = $76; { 2.6 bites ADPCM                   }
      ADPCM26r_DMA  = $77; { 2.6 bites ADPCM referencival     }
      ADPCM2_DMA    = $16; { 2 bites ADPCM                     }
      ADPCM2r_DMA   = $17; { 2 bites ADPCM referencival       }

      GetDSPver     = $e1; { A DSP-chip verziszmnak lekrdezse }

function  DSPreset:boolean;
procedure DSPwrite(DSPdata:byte);
function  DSPread:byte;
procedure DSPspeakerOn;
procedure DSPspeakerOff;
procedure Pause(Atime:word);

implementation

function DSPreset:boolean;              assembler;
asm
  mov   dx,BaseAddr
  add   dl,6   { 226 : DSP reset reg }
  mov   al,1
  out   dx,al   { dx - 226 }
  mov   cx,100
  rep lodsb
  xor   al,al
  out   dx,al   { dx - 226 }
  add   dl,8
  mov   cx,128
@wait1:
  in    al,dx   { dx - 22e }
  and   al,80h
  jnz   @test2
  loop  @wait1
  jmp   @bad
@test2:
  mov   cx,10240
  sub   dl,4    { dx - 22a }
@wait2:
  in    al,dx   { dx - 22a }
  cmp   al,0aah
  jz    @good
  loop  @wait2
@bad:
  xor   al,al
  jmp   @exit
@good:
  mov   al,1
@exit:
end;

procedure DSPwrite(DSPdata:byte);       assembler;
asm
  mov   dx,BaseAddr
  add   dl,0ch
@c1:
  in    al,dx
  and   al,80h
  jnz   @c1
  mov   al,DSPdata
  out   dx,al
end;

function DSPread:byte;                  assembler;
asm
  mov   dx,BaseAddr
  add   dl,0eh
@wait:
  in    al,dx
  and   al,80h
  jz    @wait
  sub   dl,4
  in    al,dx
end;

procedure DSPspeakerOn;
begin
 DSPwrite(SpeakerOn);
end;

procedure DSPspeakerOff;
begin
 DSPwrite(SpeakerOff);
end;

procedure Pause(Atime:word);            assembler;
asm
  mov  cx,Atime
@wait:
  nop
  nop
  loop @wait
end;

END. of unit