{$G+}
unit sb_dma;

(*****************************************************************************
*                                                                            *
*                                  D  M  A                                   *
*                                                                            *
*****************************************************************************)

interface uses dsp;

const
    Rec  = $00;
    Play = $01;

    DMA_command = $08; { Parancsregiszter                       }
    DMA_mask    = $0a; { Maszkregiszter bitenknti rsa        }
    DMA_byteptr = $0c; { A bjtmutat trlse                   }
    DMA_mode    = $0b; { Mdregiszter                           }
    DMA_status  = $08; { Sttusregiszter                        }
    DMA_req     = $09; { Krsregiszter                         }
    DMA_init    = $0d; { Inicializl regiszter                 }

    BitMasks : array[0..7] of byte=($01,$02,$04,$08,$10,$20,$40,$80);

    OldIRQofs  : word=00;
    OldIRQseg  : word=00;

const
    PageTAB : array[0..3] of byte=($87,$83,$81,$82);
    OffsTAB : array[0..3] of byte=($00,$02,$04,$06);
    SizeTAB : array[0..3] of byte=($01,$03,$05,$07);

var
    DMAdir    : byte; { Rec/Play (00/01)  }
    DMApage   : byte; { A DMA lap szma   }
    DMAoffset : word; { A DMA kezdcme   }
    DMAlength : word; { Az tvitel hossza }
    DMAmax    : word; { Maximlis tviteli hossz az aktulis lapon }
    DMAnextpg : boolean;       { Szksg van-e kvetkez tvitelre }

    { A DMAchannel rtktl fgg adatok }
    CurrentPage : word;
    CurrentOffs : word;
    CurrentSize : word;
    PlayMode    : byte;
    RecMode     : byte;
    DisDMA      : byte;
    EnDMA       : byte;

procedure StartPlay(sr:byte; song:pointer; size:word);
procedure StartRec (sr:byte; song:pointer; size:word);
procedure Refresh;
procedure ResetIRQ;

implementation

{ Bemenet:
   AX : DX  - segment : offset
  Kimenet:
   AH = page-szm
   DX = offszet
   CX = maximlis hossz ebben a page-ben
}
procedure PageMaker;                    assembler;
asm
  mov   cl,ah {cl:=AH}
  shr   cl,4
  shl   ax,4
  add   dx,ax
  mov   ah,cl
  adc   ah,00
  mov   cx,dx
  neg   cx
end;

procedure PlayDMA(page:byte; offs,size:word); forward;
procedure RecDMA (page:byte; offs,size:word); forward;

{
SoundBlaster IRQ rutin, amelyet a krtya akkor kr, amikor egy DMA
tvitel befejezdtt.
}
procedure SB_IRQ;                                              Assembler;
asm
  push  DS
  push  ES
  pusha
  mov   ax,Seg @data
  mov   DS,ax
  mov   dx,BaseAddr
  add   dx,$0e        { DSP status }
  in    al,dx         { ACK irq    }
  test  DMAnextpg,1
  jnz   @nextDMA
@endDMA:
  push  DSP_DMA_stop
  call  DSPwrite
  call  DSPspeakerOff
  call  ResetIRQ
  jmp   @endIRQ
@nextDMA:
  mov   DMAnextpg,0
  mov   al,DMApage
  inc   al
  push  ax
  push  0000
  mov   ax,DMAlength
  sub   ax,DMAmax
  push  ax
  cmp   DMAdir,Rec
  jz    @x2
  call  PlayDMA
  jmp   @endIRQ
@x2:
  call  RecDMA
@endIRQ:
  mov   al,$20
  out   $20,al        { ENDIRQ jelzs }
  popa
  pop   ES
  pop   DS
  iret
end;

{
 A SoundBlaster megszaktsi csatornjnak engedlyezse, a vektor
 rlltsa a megszaktsi rutinra.
}
procedure SetIRQ;                                              Assembler;
asm
  cli
  mov   al,IRQnumber { 2, 3, 5, 7 lehet SoundBlaster esetn }
  or    al,8
  cbw
  shl   ax,2         { *4 a vektor elhelyezse miatt }
  mov   bx,ax
  xor   ax,ax
  mov   ES,ax

  mov   dx,ES:[bx]
  mov   OldIRQofs,dx
  mov   dx,ES:[bx+2]
  mov   OldIRQseg,dx

  lea   dx,SB_IRQ
  mov   ES:[bx],dx   { ofs IRQ rut }
  mov   ES:[bx+2],CS { seg IRQ rut }
  mov   al,IRQnumber
  lea   bx,Bitmasks
  xlat               { A megfelel bitmaszkot beolvassuk a Bitmasks tmbbl }
  not   al           { A bitet nullzzuk, a tbbit 1-es lesz }
  mov   ah,al

  in    al,$21       { A megszaktsi csatornk maszkja   }
  and   al,ah        { a megadott csatornt engedlyezzk }
  out   $21,al       { A maszkot jra kirjuk             }

  sti
end;

{
 A SoundBlaster megszaktsi csatornjnak tiltsa, az eredeti vektor
 visszarsa
}
procedure ResetIRQ;                                            Assembler;
asm
  cli
  mov   al,IRQnumber { 2, 3, 5, 7 lehet SoundBlaster esetn }
  or    al,8
  cbw
  shl   ax,2         { *4 a vektor elhelyezse miatt }
  mov   bx,ax
  xor   ax,ax
  mov   ES,ax
  mov   ax,OldIRQofs
  mov   ES:[bx],ax   { old ofs IRQ rut }
  mov   ax,OldIRQseg
  mov   ES:[bx+2],ax { old seg IRQ rut }
  mov   al,IRQnumber
  lea   bx,Bitmasks
  xlat               { A megfelel bitmaszkot behozzuk }
  mov   ah,al
  in    al,$21       { Csatorna maszkok }
  or    al,ah        { A megadott csatorna tiltsa }
  out   $21,al
  xor   ax,ax
  mov   OldIRQseg,ax
  sti
end;

{ Be:
 al = page
 bx = size
 cx = offs
}
procedure SetOffsSizePage; assembler;
asm
  push  dx
  mov   dx,CurrentPage
  out   dx,al
  mov   dx,CurrentSize
  mov   al,bl
  out   dx,al
  mov   al,bh
  out   dx,al
  mov   dx,CurrentOffs
  mov   al,cl
  out   dx,al
  mov   al,ch
  out   dx,al
  pop   dx
end;

procedure PlayDMA(page:byte; offs,size:word); assembler;
asm
  dec   size
  mov   al,DisDMA
  out   DMA_mask,al
  xor   al,al
  out   DMA_byteptr,al
  mov   al,PlayMode
  out   DMA_mode,al
  mov   al,page
  mov   bx,size
  mov   cx,offs
  call  SetOffsSizePage
  mov   al,EnDMA
  out   DMA_mask,al
  push  NormalDMA_DAC
  call  DSPwrite
  push  bx
  call  DSPwrite
  mov   bl,bh
  push  bx
  call  DSPwrite
end;
procedure DPlayDMA(page:byte; offs,size:word); assembler;
asm
  dec   size
  mov   al,DisDMA
  out   DMA_mask,al
  xor   al,al
  out   DMA_byteptr,al
  mov   al,PlayMode
  out   DMA_mode,al
  mov   al,page
  mov   bx,size
  mov   cx,offs
  call  SetOffsSizePage
  mov   al,EnDMA
  out   DMA_mask,al
  push  DirectDAC
  call  DSPwrite
  push  bx
  call  DSPwrite
  mov   bl,bh
  push  bx
  call  DSPwrite
end;

procedure RecDMA(page:byte; offs,size:word); assembler;
asm
  dec   size
  mov   al,DisDMA
  out   DMA_mask,al
  xor   al,al
  out   DMA_byteptr,al
  mov   al,RecMode
  out   DMA_mode,al
  mov   al,page
  mov   bx,size
  mov   cx,offs
  call  SetOffsSizePage
  mov   al,EnDMA
  out   DMA_mask,al
  push  NormalDMA_ADC
  call  DSPwrite
  push  bx
  call  DSPwrite
  mov   bl,bh
  push  bx
  call  DSPwrite
end;

procedure StartPlay(sr:byte; song:pointer; size:word);
begin
 DSPSpeakerOff;
 DMAdir:=Play;
 DMAlength:=size;
 asm
   les  dx,song
   mov  ax,ES
   call PageMaker
   mov  DMApage,ah
   mov  DMAoffset,dx
   mov  DMAmax,cx
 end;
 DMAnextpg:=DMAmax<DMAlength;
 if not DMAnextpg then DMAmax:=DMAlength;
 DSPwrite(SetSampleRate);
 DSPwrite(sr);
 DSPspeakerOn;
 SetIRQ;
 PlayDMA(DMApage,DMAoffset,DMAmax);
end;

procedure StartRec(sr:byte; song:pointer; size:word);
begin
 DMAdir:=Rec;
 DMAlength:=size;
 asm
   les  dx,song
   mov  ax,ES
   call PageMaker
   mov  DMApage,ah
   mov  DMAoffset,dx
   mov  DMAmax,cx
 end;
 DMAnextpg:=DMAmax<DMAlength;
 if not DMAnextpg then DMAmax:=DMAlength;
 DSPwrite(SetSampleRate);
 DSPwrite(sr);
 SetIRQ;
 RecDMA(DMApage,DMAoffset,DMAmax);
end;

Procedure Refresh;
begin
 CurrentPage:=PageTAB[DMAchannel];
 CurrentOffs:=OffsTAB[DMAchannel];
 CurrentSize:=SizeTAB[DMAchannel];
 PlayMode:=$48 or DMAchannel;
 RecMode:=$44 or DMAchannel;
 DisDMA:=DMAchannel or 4;
 EnDMA:=DMAchannel and 3;
end;

BEGIN
 Refresh;
END. of unit