(* util.pas *)
unit Util;

{ ltalnos segdeljrsok }

interface

const
  NoLine     = 0;   { nincs vonal   }
  ThinLine   = 1;   { vkony vonal  }
  FatLine    = 2;   { vastag vonal  }
  DoubleLine = 3;   { dupla vonal   }
  DoubleTop  = 4;   { tetejn dupla }
  DoubleSide = 5;   { oldaln dupla }

  EnterKey = #13;
  ESC      = #27;
  UpKey    = #72;
  DownKey  = #80;
  HomeKey  = #71;
  EndKey   = #79;

type
  MenuColors = record      { men szneit tartalmaz rekord }
    FgNormal, BgNormal,
    FgBorder, BgBorder,
    FgSelect, BgSelect : byte;
  end;
  Str80      = string[80];
  ItemType   = array[1..25] of Str80;  { men ttelei }


procedure CursorOn;                         { A kurzor bekapcsolsa }
procedure CursorOff;                        { A kurzor kikapcsolsa }

procedure MenuWin(     X, Y,                { Bal fels sarok koordinti }
                       Border  : byte;      { Keret stlusa }
                       MenuCol : MenuColors;   { A men sznei }
                       Items   : ItemType;     { A menpontok }
                       ItemNum : byte;         { A menpontok szma }
                       Title   : Str80;        { A men cme }
                   var Choice  : byte);        { A vlasztott menpont }


implementation

uses Crt, Dos;

const
  NoCursor = $2000;

  Borders : array[0..5] of string[10] =
            ('          ',
             'Ŀ',
             '',
             'ͻȺ',
             '͸Գ',
             'ķӺ');

var
  OldCursor : word;

procedure SetCursor(NewCursor : word);
  var
    Reg : registers;
  begin
    with Reg do
    begin
      AH := 1;
      BH := 0;
      CX := NewCursor;
      intr($10, Reg);
    end;
  end; { procedure SetCursor }

function GetCursor : word;
  var
    Reg : registers;
  begin
    with Reg do
    begin
      AH := 3;
      BH := 0;
      intr($10, reg);
      GetCursor := CX;
    end;
  end; { function GetCursor }

procedure CursorOn;
  begin
    SetCursor(OldCursor);
  end; { procedure CursorOn }

procedure CursorOff;
  begin
    if GetCursor<>NoCursor then begin
      OldCursor := GetCursor;
      SetCursor(NoCursor);
    end;
  end; { procedure CursorOff }

procedure MenuWin;
  var
    Max, i, k : byte;
    Str       : Str80;
    Key       : char;
    Quit      : boolean;

  begin
    CursorOff;

    Max := length(Title);
    for i := 1 to ItemNum do
      if length(Items[i]) > Max then Max := length(Items[i]);
    Max := Max + 2;

    textcolor(MenuCol.FgBorder);
    textbackground(MenuCol.BgBorder);
    gotoxy(X, Y);
    Str := Borders[Border, 1];
    for i := 1 to Max do
      Str := Str + Borders[Border, 2];
    Str := Str + Borders[Border, 3];
    write(Str);

    gotoxy(X+(Max-length(Title)) div 2, Y);
    write(Borders[Border, 9]+Title+Borders[Border, 10]);

    for i := 1 to ItemNum do begin
      gotoxy(X, Y+i);
        write(Borders[Border, 8]);
      gotoxy(X+Max+1, Y+i);
        write(Borders[Border, 4]);
    end;

    gotoxy(X, Y+i+1);
    Str := Borders[Border, 7];
    for i := 1 to Max do
      Str := Str + Borders[Border, 6];
    Str := Str + Borders[Border, 5];
    write(Str);

    if Choice = 0 then Choice := 1;

    repeat
      for i := 1 to ItemNum do begin
          gotoxy(X+1, Y+i);
          if Choice = i then begin
            textcolor(MenuCol.FgSelect);
            textbackground(MenuCol.BgSelect);
            write(' '+Items[i]);
          end
          else begin
            textcolor(MenuCol.FgNormal);
            textbackground(MenuCol.BgNormal);
            write(' '+Items[i]);
          end;
          for k := 1 to Max-length(Items[i])-1 do
            write(' ');
      end;

      repeat
        Key := readkey;
        Quit := true;
        if Key = #0 then begin
          Key := readkey;
          case Key of
          UpKey   : begin
                      if Choice = 1 then Choice := ItemNum
                      else dec(Choice);
                    end;
          DownKey : begin
                      if Choice = ItemNum then Choice := 1
                      else inc(Choice);
                    end;
          HomeKey : Choice := 1;
          EndKey  : Choice := ItemNum
          else
            Quit := false;
          end;
        end
        else
          Quit := false;
          if Key = ESC then begin
            Choice := 0;
            Quit := true;
          end;
          if Key = EnterKey then Quit := true;
      until Quit;

    until Key in [EnterKey, ESC];
    CursorOn;

  end; { procedure MenuWin }

end.