(* VGAPELDA.PAS *)

{---------------------------------------------------------
 A VGA vezrl alacsonyszint kezelse beptett assembly
 alprogramokkal
 ---------------------------------------------------------}

{$G+   A 80286 utasitasok engedelyezese}

program VGA_pelda;

uses graph,crt;

const
  vgaptr   : pointer=ptr($a000,0000);

  { A VGA vezrl regiszterei: }
  addreg      = $3ce;
  setresetreg = $3cf;
      setres_reg  = $00;
      enable_reg  = $01;
      funcsel_reg = $03;
      mapsel_reg  = $04;
      mode_reg    = $05;
      bitmask_reg = $08;

  seqaddreg   = $3c4;
  resetreg    = $3c5;
     mapmaskreg  = $02;

  bitmapsize  = 38400;  { 640x480/8 bjt egy bitsk mrete }
  maxbitmaps  = 4;      { 4 bitsk -> 16 szn }

  {pontkitevsi mdok:}
  p_set = $00;
  p_and = $08;
  p_or  = $10;
  p_xor = $18;

type
 bitmapbuffer = array[1..bitmapsize] of byte;
 vgabuffer    = Array[1..maxbitmaps] of ^bitmapbuffer;

var
 yaddr:array[0..480] of word;
 xaddr:array[0..640] of word;
 mask :array[0..640] of byte;

{ A rajzolst gyorst tblzatok feltltse }
procedure Init_tables;
 var
   max_x,max_y:word;
   indx:word;
 begin
   max_x   := 640;
   max_y   := 480;
   for indx:=0 to max_y do
        yaddr[indx]:=80 * indx;
   for indx:=0 to max_x do
      begin
       asm
         mov cx,[indx]  {a byte-on beluli mask}
         mov si,cx
         and cl,7
         xor cl,7
         mov ch,1
         shl ch,cl
         mov [offset mask + si],ch
        end;
        xaddr[indx]:=indx div 8;
     end;
 end;

{ A kijellt bitsk olvassnak engedlyezse }
procedure enablemapread(map:byte);assembler;
  asm
   mov al,mapsel_reg
   mov ah,[map]
   dec ah
   mov dx,addreg
   out dx,ax
  end;

{ A kijellt bitsk rsnak engedlyezse }
procedure enablemapwrite(map:byte);assembler;
  asm
   mov al,mapmaskreg
   mov cl,[map]
   dec cl
   mov ah,1
   shl ah,cl
   mov dx,seqaddreg
   out dx,ax
  end;

{ A VGA vezrl alapllapotba vitele }
procedure resetvga;assembler;
  asm
   mov dx,addreg;

   mov al,bitmask_reg
   mov ah,0ffh
   out dx,ax

   mov ax,mode_reg
   out dx,ax

   mov ax,funcsel_reg
   out dx,ax

  mov ax,enable_reg
  out dx,ax

  mov dx,seqaddreg
  mov al,mapmaskreg
  mov ah,1111b
  out dx,ax
 end;

{ A VGA kperny adott sznnel val trlse }
procedure fillvga(color : byte); assembler;
 asm
  mov dx,addreg  {Write mode 0}
  mov al,mode_reg
  xor ah,ah
  out dx,ax

  mov al,setres_reg
  mov ah,[color] {0..15}
  shl ah,4
  add ah,[color]
  out dx,ax

  mov al,enable_reg
  mov ah,0fh
  out dx,ax

  mov ax,0a000h
  xor di,di
  mov cx,bitmapsize / 2
  mov es,ax
  cld
  rep stosw

  { a regiszterek visszalltsa }
  call resetvga
 end;

{ x,y pozciba color szn pont func mdon val kitevse }
procedure pset(x, y : word; color, func : byte); assembler;
  var
   mask:byte;
   dist:word;

  asm
   mov cx,[x]  {a bjton belli maszk}
   and cl,7
   xor cl,7
   mov ch,1
   shl ch,cl
   mov [mask],ch

   mov bx,80
   mov ax,[y]
   mul bx
   mov bx,[x]  { y * 80 }
   shr bx,3    { x shr 3 }
   add bx,ax   { bx - ofszet}
   mov [dist],bx

   mov dx,addreg

   mov al,mode_reg {read mode 0 s write mode 2}
   mov ah,2
   out dx,ax

   mov al,bitmask_reg
   mov ah,[mask]
   out dx,ax  {maszk}

   mov al,funcsel_reg
   mov ah,[func]
   out dx,ax  {rsi md (set/and/or/xor) belltsa}

   mov ax,0a000h
   mov es,ax
   mov bx,[dist]
   mov al,es:[bx]
   mov al,[color]
   mov es:[bx],al

   { a regiszterek visszalltsa }
   call resetvga
  end;

{ A kpernytartalom fjlba mentse }
procedure bsave(filename:string);
  var
   bitmap     : integer;
   screenfile : file;
  begin
   assign(screenfile,filename);
   rewrite(screenfile,1);
   for bitmap:=1 to maxbitmaps do
     begin
       enablemapread(bitmap);
       blockwrite(screenfile,vgaptr^,bitmapsize);
     end;
   resetvga;
   close(screenfile);
end;

{ A kpernytartalom visszaolvassa a fjlbl }
procedure bload(filename:string);
  var
   bitmap     : integer;
   screenfile : file;
  begin
   assign(screenfile,filename);
   reset(screenfile,1);
   for bitmap:=1 to maxbitmaps do
     begin
       enablemapwrite(bitmap);
       blockread(screenfile,vgaptr^,bitmapsize);
     end;
   resetvga;
   close(screenfile);
end;

{ A VGA kperny bekapcsolsa }
procedure VGA_on;assembler;
 asm
   mov    dx,03dah
   in     al,dx
   mov    al,0ffh
   mov    dx,03c0h
   out    dx,al
 end;


{ A VGA kperny kikapcsolsa }
procedure VGA_off;assembler;
 asm
     mov    dx,03dah
     in     al,dx
     mov    al,00h
     mov    dx,03c0h
     out    dx,al
 end;

{ Egyenes szakasz rajzolsa: (1. mdszer)
  (xb,yb)-(xe,ye) pontok kz, color sznnel,
  func mddal }
procedure line1(xb, yb, xe, ye:integer; color, func : byte); assembler;
 var
     ddx,ddy,ys:integer;
     swaped:byte;
     mask:byte;
     dist:word;
 asm
     mov ax,0a000h
     mov es,ax
     mov dx,addreg

     mov al,mode_reg {read mode 0 s write mode 2}
     mov ah,2
     out dx,ax

     mov al,funcsel_reg
     mov ah,[func]
     out dx,ax  {rsi md (set/and/or/xor) bealltsa}

     mov swaped,0

 { ddx<-abs(xe-xb) }
     mov ax,xe
     sub ax,xb
     jge @labs1
     neg ax
 @labs1:
     mov ddx,ax

 { ddy<-abs(ye-yb) }
     mov ax,ye
     sub ax,yb
     jge @labs2
     neg ax
 @labs2:
     mov ddy,ax

 {if ddy > ddx }
    mov ax,ddy
    cmp ax,ddx
    jle @lelse1

    mov swaped,1
 { xb <-> yb }
    mov  ax,xb
    xchg ax,yb
    mov  xb,ax
 { xe <-> ye }
    mov  ax,xe
    xchg ax,ye
    mov  xe,ax
 { ddx <-> ddy }
    mov  ax,ddx
    xchg ax,ddy
    mov  ddx,ax

 @lelse1:
 { if  xe < xb  }
    mov ax,xe
    cmp ax,xb
    jge @lelse2

 { xb <-> xe }
    mov  ax,xb
    xchg ax,xe
    mov  xb,ax

 { yb <-> ye }
    mov  ax,yb
    xchg ax,ye
    mov  yb,ax

 @lelse2:

 {  x:SI, y:DI,  e:BX , a:CX , b:DX }
    mov si,xb
    mov di,yb

 { ys <- sign(ye-yb) }
    mov ax,ye
    sub ax,yb
    jge @lplus
    mov ys,-1
    jmp @ltova1
 @lplus:
    mov ys,1
 @ltova1:

    mov bx,ddy
    neg bx

    mov si,xb
    dec si

 @lwhile: { while x<=xe }

   inc si
   mov cx,si
   mov dx,di

   cmp swaped,1
   jne @ltova2
   xchg cx,dx

 @ltova2:
 { pset }

   push si
   push di
   push bx
   push cx
   push dx

   mov si,cx  {a bjton belli maszk}
   and cl,7
   xor cl,7
   mov ch,1
   shl ch,cl
   {ch <- mask}

   shl dx,4
   mov ax,dx
   shl dx,2
   add ax,dx
   mov bx,si   { y * 80 }
   shr bx,3    { x shr 3 }
   add bx,ax   { bx - ofszet}
   { bx <- dist }

   mov dx,addreg
   mov al,bitmask_reg
   mov ah,ch  {mask}
   out dx,ax

   mov al,es:[bx]
   mov al,[color]
   mov es:[bx],al

   pop dx
   pop cx
   pop bx
   pop di
   pop si

   mov ax,ddy
   shl ax,1
   add bx,ax

   cmp bx,0
   jle @ltova3
   add di,ys

   mov ax,ddx
   shl ax,1
   sub bx,ax

 @ltova3:
   cmp si,xe
   jnge @lwhile

 { a regiszterek visszalltsa }
   call resetvga
 end;

{ Egyenes szakasz rajzolsa: (2. mdszer)
  (xb,yb)-(xe,ye) pontok kz, color sznnel,
  func mddal }
procedure line2(xb, yb, xe, ye:integer; color,func:byte);assembler;
var
    ddx,ddy,ys:integer;
    swaped:byte;
asm
    mov ax,0a000h
    mov es,ax
    mov dx,addreg

    mov al,mode_reg {read mode 0 s write mode 2}
    mov ah,2
    out dx,ax

    mov al,funcsel_reg
    mov ah,[func]
    out dx,ax  {rsi md (set/and/or/xor) bealltsa}

    mov swaped,0

{ ddx<-abs(xe-xb) }
    mov ax,xe
    sub ax,xb
    jge @labs1
    neg ax
@labs1:
    mov ddx,ax

{ ddy<-abs(ye-yb) }
    mov ax,ye
    sub ax,yb
    jge @labs2
    neg ax
@labs2:
    mov ddy,ax

{if ddy > ddx }
   mov ax,ddy
   cmp ax,ddx
   jle @lelse1

   mov swaped,1
{ xb <-> yb }
   mov  ax,xb
   xchg ax,yb
   mov  xb,ax
{ xe <-> ye }
   mov  ax,xe
   xchg ax,ye
   mov  xe,ax
{ ddx <-> ddy }
   mov  ax,ddx
   xchg ax,ddy
   mov  ddx,ax

@lelse1:
{ if  xe < xb  }
   mov ax,xe
   cmp ax,xb
   jge @lelse2

{ xb <-> xe }
   mov  ax,xb
   xchg ax,xe
   mov  xb,ax

{ yb <-> ye }
   mov  ax,yb
   xchg ax,ye
   mov  yb,ax

@lelse2:

{  x:SI, y:DI,  e:BX , a:CX , b:DX }
   mov si,xb
   mov di,yb

{ ys <- sign(ye-yb) }
   mov ax,ye
   sub ax,yb
   jge @lplus
   mov ys,-1
   jmp @ltova1
@lplus:
   mov ys,1
@ltova1:

   mov bx,ddy
   neg bx

   mov si,xb
   dec si

@lwhile: { while x<=xe }

  inc si
  mov cx,si
  mov dx,di

  cmp swaped,1
  jne @ltova2
  xchg cx,dx

@ltova2:
{ pset }

  push si
  push di
  push bx

  mov si,cx
  mov di,dx

  shl di,1
  mov bx,[offset yaddr + di]

  shl si,1
  add bx,[offset xaddr + si]

  mov si,cx
  mov di,dx
  mov dx,addreg
  mov al,bitmask_reg
  mov ah,[offset mask + si]
  out dx,ax
  mov dx,di

  mov al,es:[bx]
  mov al,[color]
  mov es:[bx],al

  pop bx
  pop di
  pop si

  mov ax,ddy
  shl ax,1
  add bx,ax

  cmp bx,0
  jle @ltova3
  add di,ys

  mov ax,ddx
  shl ax,1
  sub bx,ax

@ltova3:
  cmp si,xe
  jnge @lwhile

  { a regiszterek visszalltsa }
  call resetvga

end;

{ A fprogram }
var
 g1, g2, n, x, y : integer;
 maxx, maxy : word;

begin
  detectgraph(g1, g2);
  initgraph(g1,g2,'');

  if g1<>VGA then
              begin
                writeln('Csak VGA vezrlvel hasznlhat!');
              end;

  if graphresult<> grOK then halt;

  init_tables;
  maxx:= getmaxx;
  maxy:= getmaxy;

{-------------------------------------------}
  fillvga(blue);
  rectangle(0,0,getmaxx,getmaxy);
  x:=1;
  repeat
   line1(1,1,x,maxy-1,lightred,p_set);
   inc(x,8);
  until x>maxx;

  y:=1;
  repeat
   line1(1,y,maxx-1,1,yellow,p_set);
   inc(y,8);
  until y>maxy;
  readln;

{-------------------------------------------}
  fillvga(lightred);
  rectangle(0,0,getmaxx,getmaxy);
  x:=1;
  repeat
   line2(1,1,x,maxy-1,blue,p_set);
   inc(x,8);
  until x>maxx;

  y:=1;
  repeat
   line2(1,y,maxx-1,1,black,p_set);
   inc(y,8);
  until y>maxy;
  readln;

{-------------------------------------------}
  bsave('kep.dat');
  fillvga(black);
  readln;

{-------------------------------------------}
  vga_off;
  bload('kep.dat');
  vga_on;
  readln;

{-------------------------------------------}
  randomize;
  repeat
   pset(random(getmaxx)+1,random(getmaxy)+1,random(getmaxcolor)+1,p_set)
  until keypressed;
  readln;

{-------------------------------------------}
  fillvga(green);
   begin
     asm
       call getmaxx
       mov [x],ax
    end;
    y:=getmaxy;
    asm
       mov cx,[x]
       mov ax,[y]
       shr ax,1
    @xline:
       push cx
       push ax

       push cx      {x}
       push ax      {y}
       push red     {color}
       push p_set   {func}
       call pset

       pop ax
       pop cx
       loop @xline
     end;
   end;
   readln;
{-------------------------------------------}
  x:=0;
  y:=0;
  n:=32;
  repeat
   line1(x,0,x,maxy,black,p_set);
   line1(0,y,maxx,y,black,p_set);
   inc(x,n);
   inc(y,n);
  until x>maxx+1;
  readln;

{-------------------------------------------}
  x:=0;
  y:=0;
  n:=32;
  repeat
   line1(x,0,x,maxy,white,p_set);
   line1(0,y,maxx,y,white,p_set);
   inc(x,n);
   inc(y,n);
  until x>maxx+1;
  readln;

{-------------------------------------------}
  closegraph;
  restorecrtmode;
end.
